# Copyright 2014 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import copy
import os


class AnalyzerOptions(object):

    DEFAULT_BASE_TEMPLATE_FULL_IMPORT_PATH = (
        'spitfire.runtime.template.SpitfireTemplate')

    def __init__(self, **kargs):
        self.debug = False

        self.ignore_optional_whitespace = False

        # adjacent text nodes become one single node
        self.collapse_adjacent_text = False

        # generate templates with unicode() instead of str()
        self.generate_unicode = True

        # runs of whitespace characters are replace with one space
        self.normalize_whitespace = False

        # expensive dotted notations are aliased to a local variable for faster
        # lookups: write = self.buffer.write
        self.alias_invariants = False

        # when a variable defined in a block later is accessed, just use the raw
        # identifier, don't incure the cost of a resolve_placeholder call since
        # you know that this local variable will always resolve first
        self.directly_access_defined_variables = False

        # examine the 'extends' directive to see what other methods will be
        # defined on this template - that allows use to make fast calls to
        # template methods outside of the immediate file.
        self.use_dependency_analysis = False

        # if directly_access_defined_variables is working 100% correctly, you
        # can compleletely ignore the local scope, as those placeholders will
        # have been resolved at compile time. there are some complex cases where
        # there are some problems, so it is disabled for now
        self.omit_local_scope_search = False

        # once a placeholder is resolved in a given scope, cache it in a local
        # reference for faster subsequent retrieval
        self.cache_resolved_placeholders = False
        self.cache_resolved_udn_expressions = False
        # when this is enabled, $a.b.c will cache only the result of the entire
        # expression. otherwise, each subexpression will be cached separately
        self.prefer_whole_udn_expressions = False

        # Throw an exception when a udn resolution fails rather than providing a
        # default value
        self.raise_udn_exceptions = False

        # when adding an alias, detect if the alias is loop invariant and hoist
        # right there on the spot.  this has probably been superceded by
        # hoist_loop_invariant_aliases, but does have the advantage of not
        # needing another pass over the tree
        self.inline_hoist_loop_invariant_aliases = False

        # if an alias has been generated in a conditional scope and it is also
        # defined in the parent scope, hoist it above the conditional. this
        # requires a two-pass optimization on functions, which adds time and
        # complexity
        self.hoist_conditional_aliases = False
        self.hoist_loop_invariant_aliases = False

        # filtering is expensive, especially given the number of function calls
        self.cache_filtered_placeholders = False

        # generate functions compatible with Cheetah calling conventions
        self.cheetah_compatibility = False
        # use Cheetah NameMapper to resolve placeholders and UDN
        self.cheetah_cheats = False

        # the nested def doesn't make sense - unlike block, so raise an error.
        # default off for now to let people ease into it.
        self.fail_nested_defs = False

        # whether to explode on library search list accesses that are not
        # declared with #global $foo beforehand
        self.fail_library_searchlist_access = False

        # If we can skip udn resolution and instead directly access modules
        # imported via #import and #from directives.
        self.skip_import_udn_resolution = False

        # By default Spitfire will jump through hoops to resolve dot notation.
        # This flag disables this resolution and instead uses direct python
        # access. If this flag can be overriden on a per-file basis by using the
        # "#loose_resolution" directive.
        self.default_to_strict_resolution = False

        # Fully qualified import path of the base Spitfire template class.
        self.base_template_full_import_path = (
            self.DEFAULT_BASE_TEMPLATE_FULL_IMPORT_PATH)

        # Batch buffer writes as buffer.extend operations. This moves the
        # buffer operations to the end of a scope.
        self.batch_buffer_writes = False

        # Spitfire templates do not allow for the use of |raw.
        self.baked_mode = False

        # Throw compiler errors if variables are used outside of their scope.
        self.static_analysis = False

        # Disallow the use of raw in a template.
        self.no_raw = False

        # Generate line number comments in compiler output.
        self.include_sourcemap = False

        self.__dict__.update(kargs)

    def update(self, **kargs):
        self.__dict__.update(kargs)

    @classmethod
    def get_help(cls):
        return ', '.join(['[no-]' + name.replace('_', '-')
                          for name, value in vars(cls()).items()
                          if not name.startswith('__') and type(value) == bool])


default_options = AnalyzerOptions()
o1_options = copy.copy(default_options)
o1_options.collapse_adjacent_text = True

o2_options = copy.copy(o1_options)
o2_options.alias_invariants = True
o2_options.directly_access_defined_variables = True
o2_options.cache_resolved_placeholders = True
o2_options.cache_resolved_udn_expressions = True
o2_options.inline_hoist_loop_invariant_aliases = True
o2_options.use_dependency_analysis = True

o3_options = copy.copy(o2_options)
o3_options.inline_hoist_loop_invariant_aliases = False
o3_options.hoist_conditional_aliases = True
o3_options.hoist_loop_invariant_aliases = True
o3_options.cache_filtered_placeholders = True
o3_options.omit_local_scope_search = True
o3_options.batch_buffer_writes = True

optimizer_map = {
    0: default_options,
    1: o1_options,
    2: o2_options,
    3: o3_options,
}


# standard options to any compiler front-end
# @op option_parser, this object will be modified
def add_common_options(op):
    op.add_option('--preserve-optional-whitespace',
                  action='store_false',
                  default=True,
                  dest='ignore_optional_whitespace',
                  help='preserve leading whitespace before a directive')
    op.add_option('--normalize-whitespace',
                  action='store_true',
                  default=False,
                  help='normalize all runs of whitespace to one character')
    op.add_option(
        '--fail-library-searchlist-access',
        action='store_true',
        default=False,
        help='disallow searchlist accesses inside template libraries not '
        'declared with #global')
    op.add_option('--skip-import-udn-resolution',
                  action='store_true',
                  default=False,
                  help='Skip UDN resolution for imported moudles')
    op.add_option(
        '--strict-global-check',
        action='store_true',
        default=False,
        help='Throw compiler errors if display vars or extended methods'
        ' are not declared #global. Overridden by #loose_resolution')
    op.add_option('--static-analysis',
                  action='store_true',
                  default=False,
                  help='Throw compiler errors if variables are used outside '
                  'of their scope.')
    op.add_option(
        '--default-to-strict-resolution',
        action='store_true',
        default=False,
        help='Resolve dotted values in python instead of using resolve_UDN')
    op.add_option('-v', '--verbose', action='store_true', default=False)
    op.add_option('-V', '--version', action='store_true', default=False)
    op.add_option('-O', dest='optimizer_level', type='int', default=0)
    op.add_option('-o', '--output-file', dest='output_file', default=None)
    op.add_option('--xspt-mode',
                  action='store_true',
                  default=False,
                  help='enable attribute language syntax')

    op.add_option('--disable-filters',
                  dest='enable_filters',
                  action='store_false',
                  default=True)

    op.add_option('--output-directory',
                  default='',
                  action='callback',
                  callback=validate_path,
                  type='str',
                  nargs=1,
                  help='alternate directory to store compiled templates')

    op.add_option(
        '--include-path',
        default='.',
        action='callback',
        callback=validate_path,
        type='str',
        nargs=1,
        help='path to the templates hierarchy, where included files live. '
        'default: .')

    op.add_option('--base-extends-package', default=None)
    op.add_option('--extract-message-catalogue',
                  action='store_true',
                  default=False)
    op.add_option('--message-catalogue-file',
                  default=None,
                  action='callback',
                  callback=validate_path,
                  type='str',
                  nargs=1,
                  help='file to use as the message catalogue')
    op.add_option('--locale', default='')
    op.add_option('--function-registry-file',
                  default=None,
                  action='callback',
                  callback=validate_path,
                  type='str',
                  nargs=1,
                  help='file to use as the function registry')
    op.add_option('-X',
                  dest='optimizer_flags',
                  action='append',
                  default=[],
                  help=AnalyzerOptions.get_help())
    op.add_option('--tune-gc', action='store_true')
    op.add_option('--Wall',
                  action='store_true',
                  default=False,
                  dest='enable_warnings',
                  help='Show all warnings.')
    op.add_option('--Werror',
                  action='store_true',
                  default=False,
                  dest='warnings_as_errors',
                  help='Treat all warnings as errors')
    op.add_option('--compiler-stack-traces',
                  default=False,
                  action='store_true',
                  help='Get stack traces on compiler errors')
    op.add_option('--include-sourcemap',
                  default=False,
                  action='store_true',
                  help='Annotate output with sourcemap like comments.')
    op.add_option(
        '--double-assign-error',
        default=False,
        action='store_true',
        help='Throw an error if there is an unsafe double assignment.')
    op.add_option('--baked-mode',
                  default=False,
                  action='store_true',
                  help='A mode where the runtime tracks sanitization.')
    op.add_option(
        '--no-raw',
        default=False,
        action='store_true',
        help='A stricter version of spitfire where |raw is prohibited.')
    op.add_option('--base-template-full-import-path',
                  action='store',
                  type='string',
                  help='Sets the import path of the base abstract template.')


def validate_path(option, opt_str, path, parser):
    path = os.path.abspath(os.path.expanduser(path))
    setattr(parser.values, option.dest, path)



# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import copy
import os.path

from spitfire.compiler import ast
from spitfire.compiler import util
from spitfire import text


def tree_walker(node):
    yield node
    for n in node.child_nodes:
        for ng in tree_walker(n):
            yield ng


class SemanticAnalyzerError(Exception):
    pass


class MacroError(SemanticAnalyzerError):
    pass


class MacroParseError(MacroError):
    pass


i18n_function_name = 'i18n'

# This is a whitelist of nodes that are allowed at the top level in
# template libraries.
# TODO: Remove ast.TextNode once b/15314057 is fixed.
_ALLOWED_LIBRARY_NODES = (ast.TextNode, ast.ImplementsNode, ast.ImportNode,
                          ast.WhitespaceNode, ast.LooseResolutionNode,
                          ast.AllowUndeclaredGlobalsNode, ast.CommentNode,
                          ast.DefNode, ast.GlobalNode)

# This is a list of nodes that cannot exclusively make up the body of
# an if or for block.
_BLANK_NODES = (ast.WhitespaceNode, ast.CommentNode)

# convert the parse tree into something a bit more 'fat' and useful
# is this an AST? i'm not sure. it will be a tree of some sort
# this should simplify the codegen stage into a naive traversal
# even though this uses memory, i'll make a copy instead of decorating the
# original tree so i can compare the differences
# the other idea is that i can treat certain nodes as 'macros' to generate a
# few nodes that are more python-like
# additionally, there are some optimizations that are really more oriented at
# the parse tree, so i do them inline here. it's a bit split-brain, but it's
# seems easier.


class SemanticAnalyzer(object):

    def __init__(self, classname, parse_root, options, compiler):
        self.classname = classname
        self.parse_root = parse_root
        self.options = options
        self.compiler = compiler
        self.ast_root = None
        self.template = None
        self.strip_lines = False
        self.uses_raw = False
        self.base_extends_identifiers = []
        if self.compiler.base_extends_package:
            # this means that extends are supposed to all happen relative to
            # some other package - this is handy for assuring all templates
            # reference within a tree, say for localization, where each locale
            # might have its own package
            packages = self.compiler.base_extends_package.split('.')
            self.base_extends_identifiers = [
                ast.IdentifierNode(module_name) for module_name in packages
            ]

    def get_ast(self):
        ast_node_list = self.build_ast(self.parse_root)
        if len(ast_node_list) != 1:
            self.compiler.error(SemanticAnalyzerError(
                'ast must have 1 root node'))
        self.ast_root = ast_node_list[0]
        return self.ast_root

    # build an AST node list from a single parse node
    # need the parent in case we are going to delete a node
    def build_ast(self, node):
        method_name = 'analyze%s' % node.__class__.__name__
        method = getattr(self, method_name, self.default_analyze_node)
        # print method_name, node.name, node
        ast_node_list = method(node)
        try:
            if len(ast_node_list) != 1:
                return ast_node_list
        except TypeError as e:
            self.compiler.error(SemanticAnalyzerError('method: %s, result: %s' %
                                                      (method, ast_node_list)))

        # print '<', ast_node_list[0].name, ast_node_list[0]
        return ast_node_list

    def default_analyze_node(self, pnode):
        # print "default_analyze_node", type(pnode)
        return [pnode]

    # some nodes just don't need analysis
    def skip_analyze_node(self, pnode):
        return [pnode]

    analyzeIdentifierNode = skip_analyze_node
    analyzeLiteralNode = skip_analyze_node

    def analyzeTemplateNode(self, pnode):
        self.template = pnode.copy(copy_children=False)
        self.template.classname = self.classname
        # Baked mode and generate_unicode are incomaptible. This is a
        # result of SanitizedPlaceholder extending str and not unicode. A
        # possible solution is to have two classes:
        # SanitizedPlaceholderUnicode and SanitizedPlaceholderStr, each
        # that extends unicode or str. Depending on the mode, it would
        # assign SanitizedPlaceholder to be the correct class.
        if self.options.generate_unicode and self.options.baked_mode:
            self.compiler.error(
                SemanticAnalyzerError(
                    'Generate unicode is incompatible with baked mode.'),
                pos=pnode.pos)
        self.template.baked = self.options.baked_mode

        # Need to build a full list of template_methods before analyzing so we
        # can modify CallFunctionNodes as we walk the tree below.
        for child_node in tree_walker(pnode):
            if (isinstance(child_node, ast.DefNode) and
                    not isinstance(child_node, ast.MacroNode)):
                if child_node.name in self.template.template_methods:
                    self.compiler.error(
                        SemanticAnalyzerError(
                            'Redefining #def/#block %s (duplicate def in file?)'
                            % (child_node.name)),
                        pos=pnode.pos)
                self.template.template_methods.add(child_node.name)

        for pn in self.optimize_parsed_nodes(pnode.child_nodes):
            if self.template.library and not isinstance(pn,
                                                        _ALLOWED_LIBRARY_NODES):
                self.compiler.error(
                    SemanticAnalyzerError(
                        'All library code must be in a function.'),
                    pos=pn.pos)
            built_nodes = self.build_ast(pn)
            if built_nodes and not self.template.library:
                self.template.main_function.extend(built_nodes)

        if not self.uses_raw and self.template.explicitly_allow_raw:
            self.compiler.error(SemanticAnalyzerError(
                '#allow_raw directive is not needed'))

        self.template.main_function.child_nodes = self.optimize_buffer_writes(
            self.template.main_function.child_nodes)

        if self.template.extends_nodes and self.template.library:
            self.compiler.error(SemanticAnalyzerError(
                "library template can't have extends."))

        return [self.template]

    # Recursively grabs identifiers from a ast.TargetListNode, such as in a
    # ast.ForNode.
    def _getIdentifiersFromListNode(self, identifier_set, target_list_node):
        for pn in target_list_node.child_nodes:
            if isinstance(pn, ast.TargetNode):
                identifier_set.add(pn.name)
            elif isinstance(pn, ast.TargetListNode):
                self._getIdentifiersFromListNode(identifier_set, pn)

    def analyzeForNode(self, pnode):
        # If all of the children are nodes that get ignored or there are
        # no nodes, throw an error.
        if all([isinstance(pn, _BLANK_NODES) for pn in pnode.child_nodes]):
            self.compiler.error(
                SemanticAnalyzerError("can't define an empty #for loop"),
                pos=pnode.pos)

        for_node = ast.ForNode(pos=pnode.pos)

        # Backup original scope identifiers for analysis.
        template_local_scope_identifiers = set(
            self.template.local_scope_identifiers)

        self._getIdentifiersFromListNode(self.template.local_scope_identifiers,
                                         pnode.target_list)

        for pn in pnode.target_list.child_nodes:
            for_node.target_list.extend(self.build_ast(pn))
        for pn in pnode.expression_list.child_nodes:
            for_node.expression_list.extend(self.build_ast(pn))
        for pn in self.optimize_parsed_nodes(pnode.child_nodes):
            for_node.extend(self.build_ast(pn))

        for_node.child_nodes = self.optimize_buffer_writes(for_node.child_nodes)

        # Restore original scope identifiers after children have been analyzed.
        self.template.local_scope_identifiers = template_local_scope_identifiers

        return [for_node]

    def analyzeStripLinesNode(self, pnode):
        if self.strip_lines:
            self.compiler.error(
                SemanticAnalyzerError("can't nest #strip_lines"),
                pos=pnode.pos)
        self.strip_lines = True
        optimized_nodes = self.optimize_parsed_nodes(pnode.child_nodes)
        new_nodes = [self.build_ast(pn) for pn in optimized_nodes]
        self.strip_lines = False
        return self.optimize_buffer_writes(new_nodes)

    def analyzeGetUDNNode(self, pnode):
        children = pnode.getChildNodes()
        if isinstance(children[0], ast.PlaceholderNode):
            identifier = '.'.join([node.name for node in children])
            # Some modules are trusted not to need UDN resolution.
            if self._identifier_can_skip_UDN_resolution(identifier):
                expr = '%s.%s' % (identifier, pnode.name)
                return [ast.IdentifierNode(expr, pos=pnode.pos)]

        expression = self.build_ast(pnode.expression)[0]
        return [ast.GetUDNNode(expression, pnode.name, pos=pnode.pos)]

    def analyzeGetAttrNode(self, pnode):
        expression = self.build_ast(pnode.expression)[0]
        return [ast.GetAttrNode(expression, pnode.name, pos=pnode.pos)]

    def analyzeIfNode(self, pnode):
        # If all of the children are nodes that get ignored or there are
        # no nodes, throw an error.
        if all([isinstance(pn, _BLANK_NODES) for pn in pnode.child_nodes]):
            self.compiler.error(
                SemanticAnalyzerError("can't define an empty #if block"),
                pos=pnode.pos)

        if_node = ast.IfNode(pos=pnode.pos)
        if_node.else_.pos = pnode.else_.pos
        if_node.test_expression = self.build_ast(pnode.test_expression)[0]
        for pn in self.optimize_parsed_nodes(pnode.child_nodes):
            if_node.extend(self.build_ast(pn))
            if_node.child_nodes = self.optimize_buffer_writes(
                if_node.child_nodes)
        for pn in self.optimize_parsed_nodes(pnode.else_.child_nodes):
            if_node.else_.extend(self.build_ast(pn))
            if_node.else_.child_nodes = self.optimize_buffer_writes(
                if_node.else_.child_nodes)
        return [if_node]

    def analyzeFragmentNode(self, node):
        new_nodes = []
        for n in node.child_nodes:
            new_nodes.extend(self.build_ast(n))
        return new_nodes

    def analyzeArgListNode(self, pnode):
        list_node = ast.ArgListNode(pos=pnode.pos)
        for n in pnode:
            list_node.extend(self.build_ast(n))
        return [list_node]

    def analyzeTupleLiteralNode(self, pnode):
        tuple_node = ast.TupleLiteralNode(pos=pnode.pos)
        for n in pnode.child_nodes:
            tuple_node.extend(self.build_ast(n))
        return [tuple_node]

    def analyzeDictLiteralNode(self, pnode):
        dict_node = ast.DictLiteralNode(pos=pnode.pos)
        for key_node, value_node in pnode.child_nodes:
            key_value = (key_node, self.build_ast(value_node)[0])
            dict_node.child_nodes.extend([key_value])
        return [dict_node]

    def analyzeParameterNode(self, pnode):
        param = pnode
        param.default = self.build_ast(pnode.default)[0]
        return [param]

    def analyzeSliceNode(self, pnode):
        snode = pnode
        snode.expression = self.build_ast(pnode.expression)[0]
        snode.slice_expression = self.build_ast(pnode.slice_expression)[0]
        return [snode]

    # FIXME: should I move this to a directive?
    def analyzeImplementsNode(self, pnode):
        if pnode.name == 'library':
            self.template.library = True
        else:
            self.template.main_function.name = pnode.name
            self.template.implements = True
        return []

    def analyzeAllowUndeclaredGlobalsNode(self, pnode):
        self.template.allow_undeclared_globals = True
        return []

    def analyzeLooseResolutionNode(self, pnode):
        self.template.use_loose_resolution = True
        return []

    def analyzeAllowRawNode(self, pnode):
        self.template.explicitly_allow_raw = True
        return []

    def analyzeImportNode(self, pnode):
        node = ast.ImportNode(
            [self.build_ast(n)[0] for n in pnode.module_name_list],
            library=pnode.library,
            pos=pnode.pos)
        if node.library:
            self.template.library_identifiers.add('.'.join(
                node.name for node in node.module_name_list))
            node.module_name_list[0:0] = self.base_extends_identifiers

        if node not in self.template.import_nodes:
            self.template.import_nodes.append(node)
            # Modules imported via "from" are trusted to not need UDN
            # resolution.
            self.template.trusted_module_identifiers.add('.'.join(
                node.name for node in node.module_name_list))
        return []

    def analyzeExtendsNode(self, pnode):
        # an extends directive results in two fairly separate things happening
        # clone these nodes so we can modify the path struction without mangling
        # anything else
        import_node = ast.ImportNode(pnode.module_name_list[:])
        extends_node = ast.ExtendsNode(pnode.module_name_list[:])
        if type(pnode) != ast.AbsoluteExtendsNode:
            import_node.module_name_list[0:0] = self.base_extends_identifiers
            extends_node.module_name_list[0:0] = self.base_extends_identifiers

        self.analyzeImportNode(import_node)

        # actually want to reference the class within the module name
        # assume we follow the convention of module name == class name
        extends_node.module_name_list.append(extends_node.module_name_list[-1])
        self.template.extends_nodes.append(extends_node)
        return []

    analyzeAbsoluteExtendsNode = analyzeExtendsNode

    def analyzeFromNode(self, pnode):
        if pnode.library:
            self.template.library_identifiers.add(pnode.identifier.name)
            pnode.module_name_list[0:0] = self.base_extends_identifiers
        if pnode not in self.template.from_nodes:
            self.template.from_nodes.append(pnode)
            # Modules imported via "from" are trusted to not need UDN
            # resolution.
            self.template.trusted_module_identifiers.add(pnode.identifier.name)
        return []

    def analyzeTextNode(self, pnode):
        if pnode.child_nodes:
            self.compiler.error(
                SemanticAnalyzerError("ast.TextNode can't have children"),
                pos=pnode.pos)

        value = pnode.value
        if self.options.normalize_whitespace:
            value = text.normalize_whitespace(value)
        literal_node = ast.LiteralNode(value, pos=pnode.pos)
        buffer_write = ast.BufferWrite(literal_node, pos=pnode.pos)
        return [buffer_write]

    analyzeOptionalWhitespaceNode = analyzeTextNode
    analyzeWhitespaceNode = analyzeTextNode
    analyzeNewlineNode = analyzeTextNode

    # purely here for passthru and to remind me that it needs to be overridden
    def analyzeFunctionNode(self, pnode):
        return [pnode]

    def analyzeDefNode(self, pnode, allow_nesting=False):
        if (self.options.fail_nested_defs and not allow_nesting and
                not isinstance(pnode.parent, ast.TemplateNode)):
            self.compiler.error(
                SemanticAnalyzerError('nested #def directives are not allowed'),
                pos=pnode.pos)

        function = ast.FunctionNode(pnode.name, pos=pnode.pos)
        # Backup original scope identifiers for analysis.
        template_local_scope_identifiers = set(
            self.template.local_scope_identifiers)

        if pnode.parameter_list:
            # Add parameters to local template scope for static analysis in
            # children.
            self.template.local_scope_identifiers = (
                self.template.local_scope_identifiers.union(
                    [parameter.name for parameter in pnode.parameter_list]))
            function.parameter_list = self.build_ast(pnode.parameter_list)[0]

        function.parameter_list.child_nodes.insert(
            0,
            ast.ParameterNode(name='self'))

        for pn in self.optimize_parsed_nodes(pnode.child_nodes):
            function.extend(self.build_ast(pn))
        function = self.build_ast(function)[0]
        function.child_nodes = self.optimize_buffer_writes(function.child_nodes)
        self.template.append(function)

        # Restore original scope identifiers after children have been analyzed.
        self.template.local_scope_identifiers = template_local_scope_identifiers
        return []

    def analyzeBlockNode(self, pnode):
        self.analyzeDefNode(pnode, allow_nesting=True)
        function_node = ast.CallFunctionNode(pos=pnode.pos)
        function_node.expression = self.build_ast(ast.PlaceholderNode(
            pnode.name))[0]
        p = ast.PlaceholderSubstitutionNode(function_node, pos=pnode.pos)
        call_block = self.build_ast(p)
        return call_block

    def handleMacro(self, pnode, macro_function, parse_rule):
        if isinstance(pnode, ast.MacroNode):
            kargs_map = pnode.parameter_list.get_arg_map()
        elif isinstance(pnode, ast.CallFunctionNode):
            kargs_map = pnode.arg_list.get_arg_map()
        else:
            self.compiler.error(
                SemanticAnalyzerError("unexpected node type '%s' for macro" %
                                      type(pnode)),
                pos=pnode.pos)

        macro_output = macro_function(pnode, kargs_map, self.compiler)
        # fixme: bad place to import, difficult to put at the top due to
        # cyclic dependency
        try:
            if parse_rule:
                fragment_ast = util.parse(macro_output, parse_rule)
            elif isinstance(pnode, ast.MacroNode):
                fragment_ast = util.parse(macro_output, 'fragment_goal')
            elif isinstance(pnode, ast.CallFunctionNode):
                fragment_ast = util.parse(macro_output, 'rhs_expression')
        except Exception as e:
            self.compiler.error(MacroParseError(e), pos=pnode.pos)
        return self.build_ast(fragment_ast)

    def analyzeMacroNode(self, pnode):
        # fixme: better error handler
        macro_handler_name = 'macro_%s' % pnode.name
        try:
            macro_function, macro_parse_rule = self.compiler.macro_registry[
                macro_handler_name]
        except KeyError:
            self.compiler.error(
                SemanticAnalyzerError("no handler registered for '%s'" %
                                      macro_handler_name),
                pos=pnode.pos)
        try:
            temp_fragment = util.parse(pnode.value, macro_parse_rule or
                                       'fragment_goal')
        except Exception as e:
            self.compiler.error(MacroParseError(e), pos=pnode.pos)

        if not self.uses_raw:
            for child_node in tree_walker(temp_fragment):
                if (isinstance(child_node, ast.PlaceholderSubstitutionNode) and
                        'raw' in child_node.parameter_list.get_arg_map()):
                    self.uses_raw = True
                    break

        return self.handleMacro(pnode, macro_function, macro_parse_rule)

    def analyzeGlobalNode(self, pnode):
        if not isinstance(pnode.parent, ast.TemplateNode):
            self.compiler.error(
                SemanticAnalyzerError('#global must be a top-level directive.'),
                pos=pnode.pos)
        self.template.global_placeholders.add(pnode.name)
        return []

    def analyzeAttributeNode(self, pnode):
        self.template.attr_nodes.append(pnode)
        return []

    analyzeFilterAttributeNode = analyzeAttributeNode

    # note: we do a copy-thru to force analysis of the child nodes
    # this function is drastically complicated by the logic for filtering
    # basically, if you are pulling data from the search_list and writing it
    # to the output buffer, you want to do some filtering in most cases - at
    # least when you are doing web stuff.
    # there are some cases where you want to disable this to prevent double
    # escaping or to increase performance by avoiding unnecessay work.
    #
    # $test_str_function($test_dict)
    #
    # self.resolve_placeholder('test_str_function', local_vars=locals(),
    #   global_vars=_globals)(self.resolve_placeholder('test_dict',
    #   local_vars=locals(), global_vars=_globals))
    #
    # if this section is referenced inside another template-defined function the
    # data returned should not be double escaped. you can do this by forcing all
    # template functions to annotate themselves, but you have to do more
    # cumbersome checks when you are calling arbitrary functions.
    #
    # it might be reasonable to put in another node type that indicates a block
    # of data needs to be filtered.
    def analyzePlaceholderSubstitutionNode(self, pnode):
        # print ' '.join(analyzePlaceholderSubstitutionNode', pnode,
        #                pnode.parameter_list.get_arg_map())
        node_list = []
        ph_expression = self.build_ast(pnode.expression)[0]
        # If the expression contained a macro that was parsed as a
        # fragment, the expression is now a statement and can be moved
        # outside of the ast.PlaceholderSubstitutionNode.
        #
        # This is a hack to get around design decisions that were made
        # early on. It is up to the macro authors to correctly decide how
        # the macro should be parsed and the compiler should throw errors
        # if there is an odd state where nodes are somewhere unexpected.
        if isinstance(ph_expression, ast.statement_nodes):
            return [ph_expression]

        arg_map = pnode.parameter_list.get_arg_map()
        default_format_string = '%s'
        format_string = arg_map.get('format_string', default_format_string)

        skip_filter = False
        cache_forever = False
        registered_function = False
        function_has_only_literal_args = False
        never_cache = False
        if isinstance(ph_expression, ast.CallFunctionNode):
            fname = ph_expression.expression.name
            if self.compiler.registry_contains(fname):
                function_has_only_literal_args = (
                    ph_expression.arg_list and
                    not [_arg
                         for _arg in ph_expression.arg_list
                         if not isinstance(_arg, ast.LiteralNode)])
                skip_filter = self.compiler.get_registry_value(fname,
                                                               'skip_filter')
                skip_unless_baked = self.compiler.get_registry_value(
                    fname, 'skip_filter_unless_baked')
                skip_filter = skip_filter or (not self.template.baked and
                                              skip_unless_baked)
                cache_forever = self.compiler.get_registry_value(
                    fname, 'cache_forever')
                never_cache = self.compiler.get_registry_value(fname,
                                                               'never_cache')

            elif ph_expression.library_function:
                # Don't escape function calls into library templates.
                skip_filter = True

        if (self.compiler.enable_filters and
                format_string == default_format_string and
                not isinstance(ph_expression, ast.LiteralNode)):
            arg_node_map = pnode.parameter_list.get_arg_node_map()
            if 'raw' in arg_map:
                # If this is a |raw usage and the template does not allow raw,
                # raise an error.
                self.uses_raw = True
                if (self.options.no_raw and
                        not self.template.explicitly_allow_raw):
                    err_msg = ('|raw is not allowed in templates compiled '
                               'with the --no-raw flag.')
                    self.compiler.error(
                        SemanticAnalyzerError(err_msg),
                        pos=pnode.pos)
            else:
                # if we need to filter, wrap up the node and wait for further
                # analysis later on
                if skip_filter:
                    # explicitly set the filter to none here - this means we
                    # will cache expensive pseudo-filtered nodes
                    ph_expression = ast.FilterNode(ph_expression,
                                                   None,
                                                   pos=pnode.pos)
                else:
                    ph_expression = ast.FilterNode(
                        ph_expression,
                        arg_node_map.get('filter', ast.DefaultFilterFunction),
                        pos=pnode.pos)

                # if this is a literal node, we still might want to filter it
                # but the output should always be the same - so do it once and
                # cache FIXME: could fold this and apply the function at
                # compile-time
                if (not never_cache and
                    (registered_function and function_has_only_literal_args) or
                        cache_forever or 'cache' in arg_map):
                    cache_expression = ast.CacheNode(ph_expression,
                                                     pos=pnode.pos)
                    self.template.cached_identifiers.add(cache_expression)
                    node_list.append(cache_expression)
                    ph_expression = ast.IdentifierNode(cache_expression.name,
                                                       pos=pnode.pos)

        if isinstance(ph_expression, ast.LiteralNode):
            buffer_write = ast.BufferWrite(ph_expression, pos=pnode.pos)
            node_list.append(buffer_write)
        elif (self.compiler.enable_filters and
              format_string == default_format_string):
            # we are already filtering, don't bother creating a new string
            buffer_write = ast.BufferWrite(ph_expression, pos=pnode.pos)
            node_list.append(buffer_write)
        else:
            buffer_write = ast.BufferWrite(
                ast.BinOpNode('%',
                              ast.LiteralNode(format_string,
                                              pos=pnode.pos),
                              ph_expression),
                pos=pnode.pos)
            node_list.append(buffer_write)
        return node_list

    def analyzePlaceholderNode(self, pnode):
        if (self.options.fail_library_searchlist_access and
                pnode.name not in self.template.global_placeholders):
            if (self.options.strict_global_check and
                    not self.template.allow_undeclared_globals and
                (not self.template.has_identifier(pnode.name) and
                 pnode.name not in self.compiler.function_name_registry)):
                # Break compile if no #loose_resolution and variable is not
                # available in any reasonable scope.
                err_msg = ('identifier %s is unavailable and is not declared '
                           'as a #global display variable' % pnode.name)
                self.compiler.error(
                    SemanticAnalyzerError(err_msg),
                    pos=pnode.pos)
            elif self.template.library:
                # Only do placeholder resolutions for placeholders declared with
                # #global in library templates.
                identifier_node = ast.IdentifierNode(pnode.name, pos=pnode.pos)
                return [identifier_node]
        return [pnode]

    analyzeEchoNode = analyzePlaceholderNode

    def analyzeBinOpNode(self, pnode):
        n = pnode
        n.left = self.build_ast(n.left)[0]
        n.right = self.build_ast(n.right)[0]
        return [n]

    analyzeBinOpExpressionNode = analyzeBinOpNode

    def analyzeAssignNode(self, pnode):
        # Add to template's scope, which will be removed after stepping out of
        # a ast.DefNode or ast.ForNode.
        # If the left side is a ast.SliceNode, make sure the expression is an
        # ast.IdentifierNode and add the expression's name.
        if isinstance(pnode.left, ast.SliceNode):
            exp = pnode.left.expression
            if not isinstance(exp, ast.IdentifierNode):
                self.compiler.error(
                    SemanticAnalyzerError(
                        'Slice expression %s in an assign must be an identifier'
                        % exp),
                    pos=pnode.pos)
        else:
            self.template.local_scope_identifiers.add(pnode.left.name)
        return self.analyzeBinOpNode(pnode)

    def analyzeUnaryOpNode(self, pnode):
        n = pnode
        n.expression = self.build_ast(n.expression)[0]
        return [n]

    def analyzeCommentNode(self, pnode):
        return []

    def analyzeDoNode(self, pnode):
        n = pnode
        n.expression = self.build_ast(n.expression)[0]
        return [n]

    def analyzeCallFunctionNode(self, pnode):
        fn = pnode

        fname = fn.expression.name
        if self.compiler.registry_contains(fname):
            # If this is a placeholder that is in the function registry, mark it
            # as used so that in the optimizer stage, we can avoid importing
            # unused registry values.
            self.template.used_function_registry_identifiers.add(fname)

        # The fully qualified library function name iff we figure out
        # that this is calling into a library.
        library_function = None
        skip_filter = self.compiler.get_registry_value(fname,
                                                       'skip_filter')

        if isinstance(fn.expression, ast.PlaceholderNode):
            macro_handler_name = 'macro_function_%s' % fn.expression.name
            macro_data = self.compiler.macro_registry.get(macro_handler_name)
            if macro_data:
                macro_function, macro_parse_rule = macro_data
                return self.handleMacro(fn, macro_function, macro_parse_rule)
            elif fn.expression.name in self.template.template_methods:
                fn.sanitization_state = ast.SanitizedState.SANITIZED_STRING
                if self.template.library:
                    # Calling another library function from this library
                    # function.
                    library_function = fn.expression.name
            elif skip_filter:
                # If the function is marked as skip_filter in the registry, we
                # know it is sanitized.
                fn.sanitization_state = ast.SanitizedState.SANITIZED
            elif skip_filter is False:
                # If the function is marked as skip_filter=False in the
                # registry, we know it is not sanitized.
                fn.sanitization_state = ast.SanitizedState.UNSANITIZED
        elif isinstance(fn.expression, ast.GetUDNNode):
            identifier = [node.name for node in fn.expression.getChildNodes()]
            identifier = '.'.join(identifier)
            if identifier in self.template.library_identifiers:
                # Calling library functions from other templates.
                library_function = '%s.%s' % (identifier, fn.expression.name)

        if library_function:
            # Replace the placeholder node or UDN resolution with a direct
            # reference to the library function, either in another imported
            # module or here.
            fn.expression = ast.IdentifierNode(library_function, pos=pnode.pos)
            # Pass the current template instance into the library function.
            fn.arg_list.child_nodes.insert(0, ast.IdentifierNode('self'))
            # Library functions are spitfire functions so their output is
            # sanitized.
            fn.sanitization_state = ast.SanitizedState.SANITIZED_STRING
            fn.library_function = True

        fn.expression = self.build_ast(fn.expression)[0]
        fn.arg_list = self.build_ast(fn.arg_list)[0]
        return [fn]

    analyzeBufferWrite = analyzeCallFunctionNode

    def analyzeFilterNode(self, pnode):
        fn = pnode
        fn.expression = self.build_ast(fn.expression)[0]
        return [fn]

    # go over the parsed nodes and weed out the parts we don't need
    # it's easier to do this before we morph the AST to look more like python
    def optimize_parsed_nodes(self, node_list):
        optimized_nodes = []
        for n in node_list:
            # strip optional whitespace by removing the nodes
            if (self.options.ignore_optional_whitespace and
                    isinstance(n, ast.OptionalWhitespaceNode)):
                continue
            # collapse adjacent TextNodes so we are calling these buffer writes
            elif (self.options.collapse_adjacent_text and
                  isinstance(n, ast.TextNode) and len(optimized_nodes) and
                  isinstance(optimized_nodes[-1], ast.TextNode)):
                # recreate this object so it doesn't show up as whitespace
                temp_text = ast.TextNode(optimized_nodes[-1].value, pos=n.pos)
                temp_text.parent = optimized_nodes[-1].parent
                temp_text.append_text_node(n)
                optimized_nodes[-1] = temp_text
            else:
                optimized_nodes.append(n)
        # print "optimized_nodes", node_list, optimized_nodes
        return optimized_nodes

    # go over the parsed nodes and weed out the parts we don't need do this
    # after analysis as well, in case a macro generates more ast.BufferWrite
    def optimize_buffer_writes(self, node_list):
        optimized_nodes = ast.NodeList()
        for n in node_list:
            if (self.options.collapse_adjacent_text and is_text_write(n) and
                    len(optimized_nodes) and
                    is_text_write(optimized_nodes[-1])):
                optimized_nodes[-1].append_text_node(n)
            else:
                optimized_nodes.append(n)
        return optimized_nodes

    # Imported modules are trusted to not need UDN resolution.
    def _identifier_can_skip_UDN_resolution(self, identifier):
        if not self.options.skip_import_udn_resolution:
            return False
        return identifier in self.template.trusted_module_identifiers


def is_text_write(node):
    return (isinstance(node, ast.BufferWrite) and
            isinstance(node.expression, ast.LiteralNode))



# Copyright 2015 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import unittest

from spitfire.runtime import baked
try:
    from spitfire.runtime import _baked  # pylint: disable=g-import-not-at-top
except ImportError:
    _baked = None


def is_skip():
    pass


is_skip.skip_filter = True


def no_skip():
    pass


# Do not inherit from unittest.TestCase to ensure that these tests don't run.
# Add tests here and they will be run for the C and Python implementations. This
# should make sure that both implementations are equivalent.
class _BakedTest(object):

    module = None

    def setUp(self):
        self.SanitizedPlaceholder = self.module._SanitizedPlaceholder
        self.mark_as_sanitized = self.module._mark_as_sanitized
        self.runtime_mark_as_sanitized = self.module._runtime_mark_as_sanitized

    def test_make_sanitized_placeholder(self):
        self.assertEqual(str(self.SanitizedPlaceholder('bar')), 'bar')

    def test_type_sanitized_placeholder(self):
        self.assertEqual(
            type(self.SanitizedPlaceholder('bar')), self.SanitizedPlaceholder)

    def test_sanitize_number(self):
        self.assertEqual(self.mark_as_sanitized(1), 1)

    def test_sanitize_string(self):
        v = self.mark_as_sanitized('foo')
        self.assertEqual(v, self.SanitizedPlaceholder('foo'))
        self.assertEqual(type(v), self.SanitizedPlaceholder)

    def test_runtime_sanitize_false(self):
        v = self.runtime_mark_as_sanitized('foo', no_skip)
        self.assertEqual(v, 'foo')
        self.assertEqual(type(v), str)

    def test_runtime_sanitize_true_not_string(self):
        v = self.runtime_mark_as_sanitized(1, is_skip)
        self.assertEqual(v, 1)

    def test_runtime_sanitize_true(self):
        v = self.runtime_mark_as_sanitized('foo', is_skip)
        self.assertEqual(v, 'foo')
        self.assertEqual(type(v), self.SanitizedPlaceholder)

    def test_add_sanitized(self):
        a = self.SanitizedPlaceholder('foo')
        b = self.SanitizedPlaceholder('bar')
        v = a + b
        self.assertEqual(v, 'foobar')
        self.assertEqual(type(v), self.SanitizedPlaceholder)

    def test_add_not_sanitized(self):
        a = self.SanitizedPlaceholder('foo')
        b = 'bar'
        v = a + b
        self.assertEqual(v, 'foobar')
        self.assertEqual(type(v), str)

    def test_add_reverse_not_sanitized(self):
        a = self.SanitizedPlaceholder('foo')
        b = 'bar'
        v = b + a
        self.assertEqual(v, 'barfoo')
        self.assertEqual(type(v), str)

    def test_mod_sanitized(self):
        a = self.SanitizedPlaceholder('foo%s')
        b = self.SanitizedPlaceholder('bar')
        v = a % b
        self.assertEqual(v, 'foobar')
        self.assertEqual(type(v), self.SanitizedPlaceholder)

    def test_mod_not_sanitized(self):
        a = self.SanitizedPlaceholder('foo%s')
        b = 'bar'
        v = a % b
        self.assertEqual(v, 'foobar')
        self.assertEqual(type(v), str)

    def test_mod_dict_not_sanitized(self):
        a = self.SanitizedPlaceholder('foo%(b)s')
        b = 'bar'
        v = a % vars()
        self.assertEqual(v, 'foobar')
        self.assertEqual(type(v), str)


class TestBakedPy(_BakedTest, unittest.TestCase):
    module = baked


if _baked is not None:

    class TestBakedC(_BakedTest, unittest.TestCase):
        module = _baked


if __name__ == '__main__':
    unittest.main()



# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# an 'abstract' base class for a template, seems like a good idea for now

import io as StringIO

from spitfire import runtime
from spitfire.runtime import baked
from spitfire.runtime import filters
from spitfire.runtime import udn

try:
    from spitfire.runtime import _template  # pylint: disable=g-import-not-at-top
except ImportError:
    _template = None


# NOTE: in some instances, this is faster than using cStringIO
# this is slightly counter intuitive and probably means there is more here than
# meets the eye.
class BufferIO(list):
    write = list.append

    def getvalue(self):
        return ''.join(self)


class _BaseSpitfireTemplate(object):

    # filter_function checks if the value should be filtered. If it is a
    # SanitizedPlaceholder or the placeholder_function has a skip_filter
    # annotation, there is no need to filter. Otherwise, call
    # self._filter_function.
    def filter_function(self, value, placeholder_function=None):
        """Checks if the value should be filtered.

        If it is a SanitizedPlaceholder or the placeholder_function has
        a skip_filter annotation, there is no need to filter. Otherwise, call
        self._filter_function.

        Args:
          value: The value that may need to be filtered.
          placeholder_function: If present and annotated, do not filter.

        Returns:
          value, filtered if necessary.
        """
        if isinstance(value, baked.SanitizedPlaceholder):
            return value
        elif (placeholder_function is not None and
              getattr(placeholder_function, 'skip_filter', False)):
            return value
        else:
            return self._filter_function(value)


def get_spitfire_template_class(prefer_c_extension=True):
    """Returns an appropriate SpitfireTemplate class.

    Args:
      prefer_c_extension: If set True and _template loaded properly, use the
        C extension's baseclass implementation.

    Returns:
      A SpitfireTemplate class with an appropriate base class.
    """
    if prefer_c_extension and _template is not None:
        baseclass = _template.BaseSpitfireTemplate
    else:
        baseclass = _BaseSpitfireTemplate

    class _SpitfireTemplate(baseclass):
        # store a reference to the filter function - this is tricky because of
        # some python stuff. filter functions look like this:
        #
        # def filter_function(template_instance, value):
        #
        # when this is assigned to a template instance, accessing this name
        # binds the function to the current instance. using the name
        # 'template_instance' to indicate that these functions aren't really
        # related to the template.
        _filter_function = staticmethod(filters.simple_str_filter)
        repeat = None
        placeholder_cache = None

        def __init__(self,
                     search_list=None,
                     default_filter=None,
                     use_placeholder_cache=False):
            # use_placeholder_cache - cache the values returned from the
            # search_list?   The cached values will live for the lifetime of
            # this object.
            self.search_list = search_list
            if use_placeholder_cache:
                self.placeholder_cache = {}
            if default_filter is not None:
                self._filter_function = default_filter

            # FIXME: repeater support is not needed most of the time, just
            # disable it for the time being
            # self.repeat = spitfire.runtime.repeater.RepeatTracker()

        def get_var(self, name, default=None):
            return udn.resolve_from_search_list(self.search_list, name, default)

        def has_var(self, name):
            var = self.get_var(name, default=runtime.UnresolvedPlaceholder)
            return var is not runtime.UnresolvedPlaceholder

        @staticmethod
        def new_buffer():
            return BufferIO()

    return _SpitfireTemplate


SpitfireTemplate = get_spitfire_template_class()


def template_method(function):
    function.template_method = True
    function.skip_filter = True
    return function
